#include "pic24_all.h"

// Copyright (c) Hari N. Nair, Bangalore 2010

// If you do not switch off the unit manually, it will go to sleep mode automatically if it does not 
// detect significant vertical movement in a period of 15 minutes.

// With a new 3V CR2032 battery, observed circuit current drain in operating mode was ~ 675uA.
// Sleep current < 10uA is possible  with the use of good quality SMD ceramic capacitors (low leakage
// current)

// Low power FRC clock 500kHz. CLKDIV set to 0 => Divide by 1 => output clock is 500kHz. 
// MCU instruction clock is 500kHz/2 = 250kHz
// Brown-out reset voltage is set to 1.8V

_FBS( BSS_OFF & BWRP_OFF )
_FGS( GSS0_OFF & GWRP_OFF )
_FOSCSEL( FNOSC_LPFRCDIV & IESO_OFF )
_FOSC( FCKSM_CSECMD & POSCFREQ_100KHZ & OSCIOFNC_ON & POSCMOD_NONE)
_FWDT( FWDTEN_OFF & WINDIS_OFF & FWPSA_PR32 & WDTPS_PS1 )
_FPOR( MCLRE_ON & BORV_1_8V & AI2C1SEL_PRI & PWRTEN_OFF & BOREN_BOR3)
_FICD( BKBUG_OFF & COE_OFF & ICS_PGx3 )
_FDS( DSWDTEN_OFF & DSBOR_OFF & RTCCREF_OFF & DSWDTLPRC_OFF & DSWDTPS_DSWDTPS_0 )


// if DEBUG is enabled, the UART is enabled and debug messages are printed to the UART terminal
// using UART2 TX pin (this does double duty with the PGD pin). The messages can be displayed using
// the Microchip Pickit2 UART window, in the drop-down baudrate selection, select "custom baudrate"
// and set it to 8928.

// The debug messages include the sensor calibration coefficients and the cruise altitude and pressure.

//#define DEBUG

#ifdef DEBUG
   #include <stdio.h>
#endif

// Sampling frequency is approximately 20 temperature compensated pressure readings per second
#define SENSOR_SAMPLES_PER_SEC        20

// 3 second averaging window is sufficient to average out most of the sensor noise, but
// still allow acceptable response delay to changing altitude
#define SNS_NUM_Z_SAMPLES            60

#define MCU_STATE_SLEEP  1010
#define MCU_STATE_VARIO  2020

// pressure sensor sampling states
#define SENSOR_READ_TEMPERATURE 		11
#define SENSOR_READ_PRESSURE			22

#define I2C_ID_BMP085         ((u8)0xEE)

#define BEEP_DISCR_THRESHOLD    30L

// max climbrate that is resolved by (beyond this, no change in the audio tone)
#define VARIO_MAX_CPS         1000L
// more audio discrimination for climbrates up to 400cm/s
#define VARIO_XOVER_CPS       400L

// change this based on the frequency bandwidth of the piezo transducer
// this is good for the Kingstate KPEG006 piezo
#define VARIO_MAX_FREQHZ      4800L
#define VARIO_XOVER_FREQHZ    2000L
#define VARIO_MIN_FREQHZ      400L

#define VARIO_SINK_FREQHZ     1500L

#define KLPF_ACCEL   			25
#define KLPF_CLIMBRATE   		191

#define SINKTONE_CPS_DFLT     -300L
#define SINKTONE_CPS_MIN      -400L
#define SINKTONE_CPS_MAX      -100L

#define CLIMBTONE_CPS_MAX     60L
#define CLIMBTONE_CPS_DFLT    15L
#define CLIMBTONE_CPS_MIN     -30L  

// sw will automatically put the mcu in sleep mode to conserve battery power
// if no significant vertical movement is detected for 15 minutes 
#define SLEEP_TIMEOUT_MINUTES             15
#define SLEEP_THRESHOLD_CPS				  50

#define AUDIO_CHIRP

#define TONE_LO_FREQHZ  250L
#define TONE_MID_FREQHZ 1000L
#define TONE_HI_FREQHZ	4000L

// AUDIO_MODE_SPEED : audio beep/chirp base frequency and repeat rate are both proportional to vertical climbrate
// AUDIO_MODE_SPEED_ACCEL : audio beep/chirp base frequency is proportional to vertical climbrate, repeat rate 
//                   is proportional to vertical acceleration/deceleration
// option is determined by shorting J1 to ground, or leaving it open
#define AUDIO_MODE_SPEED_ACCEL	0
#define AUDIO_MODE_SPEED		1

// option is determined by shorting J2 to ground, or leaving it open
#define FILTER_MODE_ADAPTIVE		0
#define FILTER_MODE_PLAIN			1

// pin 6 is connected to pin 6 of the pickit2 connector, debug signals on this pin
// are observed using the Pickit2 Logic Analyzer window to profile the
// pressure sampling and computation timing to validate that the unit is 
// operating correctly
#define gpioDbg         _LATB2 // pin 6 toggled to profile code timing
#define gpioAudioMode   _RA2  // pin 7 J1
#define gpioFilterMode  _RB4  // pin 9 J2
#define gpioSnsEoc      _RA4  // pin 10 bmp085 end of conversion 

int gnSensorState;
int gnSmpCnt;

// Use BMP085 oversampling mode (sensor averages series of ADC samples internally)
s32 gnSensorOversampling = 3L;

// Timer, variables used to generate the automatic timeout sleep 
volatile u32 gnTimeTick;
volatile int gnSecTick;
volatile int gnElapsedHrs;
volatile int gnElapsedMin;
volatile int gnElapsedSec;
volatile int gbTmr3Flag;
volatile int gbTmr1Flag;

s32 gZBuf[SNS_NUM_Z_SAMPLES];
s32 gnAltAvgM; 
s32 gnAltAvgCm;
s32 gnCps;
s32 gnSum_t;
s32 gnSum_t2;
s32 gnDen;
s32 gnPa;

// internal BMP085 intermediate data used in computation of pressure and temperature.
// Note that the BMP085 always seems to be computing temperature a few degrees C higher
// than the actual ambient temperature.
u32 gnTraw;
u32 gnPraw;
s16 gnTempC;

s16 gnBeepPeriodTicks = 0;
s16 gnBeepTransition;
s32 gnBeepCps = 0;
s32 gnVarioCps = 0;
s32 gnFreqHz;

s32 gnSinkToneCps  = SINKTONE_CPS_DFLT;
s32 gnClimbToneCps = CLIMBTONE_CPS_DFLT;
s32 gnKLpf = KLPF_CLIMBRATE;
s16 gnAccel = 0;

// BMP085 calibration coefficients
s16 gAC1;
s16 gAC2;
s16 gAC3;
u16 gAC4;
u16 gAC5;
u16 gAC6;
s16 gB1;
s16 gB2;
s32 gB5;
s16 gMb;
s16 gMc;
s16 gMd;

// temp buffer used to print debug messages to the UART. Can be seen on the pickit2
// UART terminal window
 
#ifdef DEBUG
   char gszBuf[60];
#endif

void audio_Config(void);
void audio_SetFreq(s32 nFreqHz);
void audio_Beep(s32 nCps);
void audio_Tone(s32 freqHz);

void sns_TriggerPressureSample(void);
void sns_TriggerTemperatureSample(void);
u32  sns_ReadPressureSample(void);
u32  sns_ReadTemperatureSample(void);
s16  sns_CalcTemperatureCx10(void);
s32  sns_CalcPressurePa(void);
void sns_DeltaCmPerSec(void);
void sns_Init(void);
void sns_InitData(void);
void sns_InitWindowBuffers(void);
void sns_ReadCoeffs(void);
void sns_SampleContinuous(void);
s32  sns_Pa2Cm(s32 pa);
void sns_AverageAltitude(void);
void sns_LPFClimbRate(void);
void sns_LPFAccel(void);
s32  sns_AdaptKLpf(void);

void mcu_Init(void);
static inline void mcu_Sleep();
void mcu_PeriphConfig(void);
void mcu_PeriphDisable(void);

void cmp_Config(void);
void tmr1_Config(void);
void tmr1_BusyWaitMs(u32 mS);
void tmr1_Start(void);
int  tmr1_ElapsedTimeMs(void);
void tmr1_FlagWaitMs(u32 mS);
void tmr2_Config(void);
void tmr3_Config(void);

#ifdef DEBUG
   void uart_Config(void);
   void uart_Print(char *szBuf);
   void sns_BoschCodeTest(void);
#endif 

void i2c_ReadN(u8  addr,u8* pdata, u16  cnt);
void i2c_Read2(u8  addr,u8* pd1, u8* pd2);
void i2c_Read1 (u8  addr,u8* pd1);
void i2c_WriteN(u8  addr,u8* pdata, u16  cnt);
void i2c_Write2(u8 addr,u8 d1, u8 d2);
void i2c_Write1(u8 addr,u8 d1);
u8   i2c_Get(u8 ack2Send);
u8   i2c_PutNoAckCheck(u8 uval);
void i2c_Put(u8 val);
void i2c_Stop(void);
void i2c_ReStart(void);
void i2c_Start(void);
void i2c_Config(void);
int  i2c_RcvBuf(u8 devId, u8 addr, int nBytes, u8* pBuf);
int  i2c_XmtByte(u8 devId, u8 addr, u8 data);


const char szCopyright[] = "Copyright (c) Hari N. Nair, Bangalore, 2010";
const char szVersion[] = "Ver 0.97";

int main(void) {
   mcu_Init();
#ifdef DEBUG
   tmr1_BusyWaitMs(3000);
#endif
   sns_Init();
   sns_InitData();

   while (1) {
      if (gnElapsedMin >= SLEEP_TIMEOUT_MINUTES) {
#ifdef DEBUG
         uart_Print("15min no movement timeout sleep\r\n");
#endif
         mcu_Sleep();
         } 
      else {
         sns_SampleContinuous();           
         }
	  }
   return 0;
   }


static inline void mcu_Sleep() {
#ifdef DEBUG
   uart_Print("Going to sleep\r\n");
#endif
   tmr1_Config();
   audio_Tone(TONE_HI_FREQHZ);
   audio_Tone(TONE_MID_FREQHZ);
   audio_Tone(TONE_LO_FREQHZ);
   audio_SetFreq(0);
   gpioDbg = 0;
   mcu_PeriphDisable();
   Nop();Nop();Nop();
   Sleep();
   }



void mcu_Init(void) {
   AD1PCFG = 0xFFFF;            // set all analog-input enabled pins (ANx) to digital mode
   CLKDIVbits.RCDIV = 0;  		// LPFRC oscillator 500kHz divided by 1
   CONFIG_RB2_AS_DIG_OUTPUT();  // 6 : gpioDbg
   CONFIG_RA2_AS_DIG_INPUT();   // 7 : J1
   ENABLE_RA2_PULLUP();
   CONFIG_RB4_AS_DIG_INPUT();   // 9 : J2
   ENABLE_RB4_PULLUP();
   CONFIG_RA4_AS_DIG_INPUT();   // 10 : BMP085 End of Conversion
   CONFIG_RA6_AS_DIG_OUTPUT();  // 14 : Audio output
   mcu_PeriphConfig();
   }

void mcu_PeriphConfig(void) {
   i2c_Config();
   audio_Config();
   tmr1_Config();
   tmr3_Config();
   cmp_Config();
   gpioDbg = 0;
#ifdef DEBUG
   uart_Config();
#endif
   }

void mcu_PeriphDisable(void) {
   T1CONbits.TON = 0;    // disable tmr1
   _T1IF = 0;            //clear interrupt flag
   _T1IE = 0;            //disable the interrupt
   T2CONbits.TON = 0;    // disable tmr2
   OC1CONbits.OCM = 0; // disable ocm
   _T3IF = 0;            //clear interrupt flag
   _T3IE = 0;            // disable interrupt
   T3CONbits.TON = 0;    // disable tmr3
#ifdef DEBUG
   U2MODEbits.UARTEN = 0; // disable uart2
   // clear IF flags 
   IFS1bits.U2RXIF = 0;
   IFS1bits.U2TXIF = 0;
   // disable interrupt 
   IEC1bits.U2RXIE = 0;
   IEC1bits.U2TXIE = 0;
#endif
   I2C1CONbits.I2CEN = 0; // disable i2c

   CM1CONbits.COE = 0;  // comparator output disabled
   CM1CONbits.CEN = 0;  // comparator disable

   // config i/o pins as digital inputs to minimize power 
   CONFIG_RA0_AS_DIG_INPUT();
   CONFIG_RA1_AS_DIG_INPUT();
   CONFIG_RB0_AS_DIG_INPUT();
   CONFIG_RB1_AS_DIG_INPUT();
   CONFIG_RB8_AS_DIG_INPUT();
   CONFIG_RB9_AS_DIG_INPUT();
   _LATB8 = 1;
   _LATB9 = 1;
   CONFIG_RB2_AS_DIG_INPUT();
   CONFIG_RA2_AS_DIG_INPUT();
   CONFIG_RA3_AS_DIG_INPUT();
   CONFIG_RB4_AS_DIG_INPUT();
   CONFIG_RA4_AS_DIG_INPUT();

   CONFIG_RB15_AS_DIG_INPUT();
   CONFIG_RB14_AS_DIG_INPUT();
   CONFIG_RB13_AS_DIG_INPUT();
   CONFIG_RB12_AS_DIG_INPUT();
   CONFIG_RA6_AS_DIG_INPUT();  
   CONFIG_RB7_AS_DIG_INPUT();
   }   

void cmp_Config(void) {
   CM1CONbits.COE = 1;  // output goes to pin 17
   CM1CONbits.CPOL = 1; // inverted
   CM1CONbits.CLPWR = 0; // no operation in low power mode
   CM1CONbits.EVPOL = 0; // no interrupt/event
   CM1CONbits.CREF = 0; // + input connected to C1INA (pin 8)
   CM1CONbits.CCH = 3; // - input connected to VBG/2
   CM1CONbits.CEN = 1;  // enable
 
   Nop();Nop();Nop();Nop();Nop();
   _C1EVT = 0; //clear C1 event flag
   _CMIF = 0;  //clear interrupt flag
  }

// TMR2 used for generating audio tone square wave 
void tmr2_Config(void) {
   T2CONbits.T32 = 0; // 16bit timer
   T2CONbits.TCS = 0; //  select internal clock 
   T2CONbits.TGATE = 0; // gate disabled
   T2CONbits.TSIDL = 0; // continue in idle mode
   T2CONbits.TCKPS1 = 0; // Prescaler 1:8
   T2CONbits.TCKPS0 = 1; // Prescaler 1:8
   T2CONbits.TON = 1; // start tmr2
   }

void audio_Config(void) {
   tmr2_Config();
   OC1CONbits.OCTSEL = 0; // use TMR2
   }

void audio_SetFreq(s32 nFreqHz) {
   u16 tmrTicks;
   OC1CONbits.OCM = 0; // disable ocm
   T2CONbits.TON = 0; // stop tmr2
   TMR2 = 0;
   if (nFreqHz) {
      tmrTicks = (u16)((FCY>>3)/nFreqHz);
      OC1R = (tmrTicks >> 1); // square wave
      OC1RS = tmrTicks;
      PR2 = tmrTicks;
      T2CONbits.TON = 1; // start tmr2
      OC1CONbits.OCM = 5; // continuous pulse mode
      }
    else { // dc 0V
      OC1R = 50;
      OC1RS = 100;
      PR2 = 40;
      }
   }

void audio_Tone(s32 freqHz) {
   audio_SetFreq(freqHz);
   tmr1_BusyWaitMs(300);
   audio_SetFreq(0);
   }

//  +---+---+---+ BeepPeriod +----+----+----+----+----+----+----+
//                                +----+----+--BeepTransition---+
//
//  Each tick is one pressure value sampling cycle  (approximately 1/15 seconds).
//  A counter is initialized to the value BeepPeriod at the start of the tone generation. 
//  Each tick, the counter is decremented. When it reaches the value BeepTransition, the tone
//  is shut off. The whole cycle repeats when the counter reaches zero.

// use this table for changing beep repeat rate based on vertical acceleration/deceleration
const s16 gnModeABeepTbl[5][2] = {
{8,1}, 
{6,1},
{4,1}, 
{3,1}, 
{2,1} 
};

void sns_LPFAccel(void)   {
   	s16 tmp;
	static s16 nAccelx32 = 0;
   	tmp = nAccelx32*KLPF_ACCEL + (gnAccel<<5)*(32-KLPF_ACCEL);
   	nAccelx32 = (tmp >= 0 ? ((tmp + 16)>>5) : ((tmp - 16)>>5));
   	gnAccel = (nAccelx32>>5);
   	}  

void sns_LPFClimbRate(void)   {
   	s32 tmp;
	static s32 nCpsx256 = 0;
   	tmp = nCpsx256*gnKLpf + (gnCps<<8)*(256-gnKLpf);
   	nCpsx256 = (tmp >= 0 ? ((tmp + 128)>>8) : ((tmp - 128)>>8));
   	gnCps = (nCpsx256>>8);
   	}  

s32 sns_AdaptKLpf(void) {
    s32	k,ac;
    if (gnCps < -10) k = 0;
    else {
        ac = ABS(gnAccel);
	    k = ((gnAccel >= 0)? 232 - (232*ac)/240 : 232 - (232*ac)/80);
		}
    CLAMP(k,0,232);
    return k;
	}

// use this table for changing beep repeat rate based on vertical speed
const s16 gnModeSBeepTbl[20][2] = {
{8,1}, 
{7,1},
{6,1}, 
{6,1}, 
{5,1}, 
{5,1}, 
{4,1}, 
{4,1}, 
{3,1}, 
{3,1}, 
{2,1},
{2,1}, 
{2,1}, 
{2,1}, 
{2,1}, 
{2,1}, 
{2,1}, 
{2,1}, 
{2,1},
{2,1}  
};


void audio_Beep(s32 nCps) {
   static s32 nFreqHz;
   int nRange;

   if ((gnBeepPeriodTicks <= 0) || (ABS(nCps-gnVarioCps) > BEEP_DISCR_THRESHOLD) || ((nCps > 0) && (gnVarioCps <= 0)) ) {
      gnVarioCps = nCps;
      // if sinking much faster than glider sink rate, generate continuous tone alarm
      if (gnVarioCps <= gnSinkToneCps) {
         gnBeepCps = gnVarioCps;	
         nFreqHz = VARIO_SINK_FREQHZ;
         audio_SetFreq(nFreqHz);
         gnBeepPeriodTicks = 0;
         gnBeepTransition  = 0;
         }
      //if climbing, generate beeps
      else {
         gnBeepCps = gnVarioCps - gnClimbToneCps;
         if (gnBeepCps > 0) {
            if (gnBeepCps > VARIO_MAX_CPS) {
				gnBeepCps = VARIO_MAX_CPS;
				}
		 	if (gpioAudioMode == AUDIO_MODE_SPEED_ACCEL) {
            	nRange = ABS(gnAccel)/7;
            	if (nRange > 4) nRange = 4;
            	gnBeepPeriodTicks = gnModeABeepTbl[nRange][0];
            	gnBeepTransition = gnModeABeepTbl[nRange][1];
				}
			else { // gpioAudioMode == AUDIO_MODE_SPEED
            	nRange = gnBeepCps/50;
            	if (nRange > 19) nRange = 19;
            	gnBeepPeriodTicks = gnModeSBeepTbl[nRange][0];
            	gnBeepTransition = gnModeSBeepTbl[nRange][1];
				}	
         	if (gnBeepCps > VARIO_XOVER_CPS) {
               nFreqHz = VARIO_XOVER_FREQHZ + ((gnBeepCps - VARIO_XOVER_CPS)*(VARIO_MAX_FREQHZ - VARIO_XOVER_FREQHZ))/(VARIO_MAX_CPS - VARIO_XOVER_CPS);
               }
            else {
               nFreqHz = VARIO_MIN_FREQHZ + (gnBeepCps*(VARIO_XOVER_FREQHZ - VARIO_MIN_FREQHZ))/VARIO_XOVER_CPS;
               }
            audio_SetFreq(nFreqHz);
            }
      // if  within normal glider sink rate, be quiet
         else{
            gnBeepPeriodTicks = 0;
            gnBeepTransition  = 0;
            nFreqHz = 0;
            audio_SetFreq(nFreqHz);
            }
         }
      }
   else{ 
      gnBeepPeriodTicks--;
      if (gnBeepPeriodTicks <= gnBeepTransition){ // shut off tone
         nFreqHz = 0;
         audio_SetFreq(nFreqHz);
         }
#ifdef AUDIO_CHIRP
	  else {
		nFreqHz += 10;
        audio_SetFreq(nFreqHz);
		}
#endif
	  }
   }

void sns_Init(void) {
   gnCps = 0;
   gnPa = 0;
   gnAltAvgM = 0;
   gnAltAvgCm = 0;
   // following parameters used in climb/sink rate calculation need to be computed only when 
   // the number of window samples changes. 
   gnSum_t = -((s32)SNS_NUM_Z_SAMPLES * (SNS_NUM_Z_SAMPLES-1L))/2L;
   gnSum_t2 = ((s32)SNS_NUM_Z_SAMPLES * (SNS_NUM_Z_SAMPLES-1L)*(2L*SNS_NUM_Z_SAMPLES-1L))/6L;
   gnDen = ((s32)SNS_NUM_Z_SAMPLES*gnSum_t2) - (gnSum_t*gnSum_t);

   sns_ReadCoeffs();
#ifdef DEBUG
  // sns_BoschCodeTest();
#endif
   }


void sns_ReadCoeffs(void) {
   u8 b[22];
   i2c_RcvBuf(I2C_ID_BMP085,0xAA,22,b);

   gAC1 = (s16)((((u16)b[0])<<8) | (u16)b[1]);
   gAC2 = (s16)((((u16)b[2])<<8) | (u16)b[3]);
   gAC3 = (s16)((((u16)b[4])<<8) | (u16)b[5]);

   gAC4 = (((u16)b[6])<<8) | (u16)b[7];
   gAC5 = (((u16)b[8])<<8) | (u16)b[9];
   gAC6 = (((u16)b[10])<<8) | (u16)b[11];

   gB1 = (s16)((((u16)b[12])<<8) | (u16)b[13]);
   gB2 = (s16)((((u16)b[14])<<8) | (u16)b[15]);
   gMb = 0;
   gMc = (s16)((((u16)b[18])<<8) | (u16)b[19]);
   gMd = (s16)((((u16)b[20])<<8) | (u16)b[21]);
#ifdef DEBUG
   uart_Print("Calibration coefficients:\r\n");
   sprintf(gszBuf,"AC1 %d\r\nAC2 %d\r\nAC3 %d\r\n",gAC1, gAC2, gAC3);
   uart_Print(gszBuf);
   sprintf(gszBuf,"AC4 %u\r\nAC5 %u\r\nAC6 %u\r\n", gAC4, gAC5, gAC6);
   uart_Print(gszBuf);
   sprintf(gszBuf,"B1 %d\r\nB2 %d\r\n",gB1, gB2);
   uart_Print(gszBuf);
   sprintf(gszBuf,"MB %d\r\nMC %d\r\nMD %d\r\n",gMb, gMc, gMd);
   uart_Print(gszBuf);

   uart_Print("Example values provided by Bosch : \r\n");
   uart_Print("AC1 11663\r\nAC2 -826\r\nAC3 -14324\r\n");
   uart_Print("AC4 32514\r\nAC5 25609\r\nAC6 16273\r\n");
   uart_Print("B1 5498\r\nB2 38\r\n");
   uart_Print("MB 0\r\nMC -11075\r\nMD 2432\r\n");
#endif
   }

void sns_InitData(void) {
   sns_InitWindowBuffers();
   sns_TriggerTemperatureSample();
   gnSensorState = SENSOR_READ_TEMPERATURE;
   tmr1_FlagWaitMs(25);
   }

void sns_SampleContinuous(void) {
   s16 t;   
   if (!gbTmr1Flag) return;
   if (gnSensorState == SENSOR_READ_TEMPERATURE) {
      gnTraw = sns_ReadTemperatureSample();
      sns_TriggerPressureSample();
      tmr1_FlagWaitMs(30);
      gpioDbg = 1; // turn on the debug pulse for timing the critical computation
      t = sns_CalcTemperatureCx10();
      gnTempC = (t >= 0?  (t+5)/10 : (t-5)/10);
      gnPa = sns_CalcPressurePa();
      gZBuf[gnSmpCnt] = sns_Pa2Cm(gnPa);
      sns_AverageAltitude();
   	  sns_DeltaCmPerSec();
	  if (gpioFilterMode == FILTER_MODE_ADAPTIVE) {
		 gnKLpf = sns_AdaptKLpf();
		 }
      else {
         gnKLpf = KLPF_CLIMBRATE;
		 }
	  sns_LPFClimbRate();
   	  gnSmpCnt++;
   	  if (gnSmpCnt == SNS_NUM_Z_SAMPLES) {
       	 gnSmpCnt = 0;
       	 }
      audio_Beep(gnCps);
      // reset the sleep timeout counter if there is significant vertical movement
      if (ABS(gnCps) > SLEEP_THRESHOLD_CPS) {
		 gnElapsedHrs = 0;
         gnElapsedMin = 0;
         gnElapsedSec = 0;
         gnSecTick = 0;
 		 }
      gpioDbg = 0;
	  gnSensorState = SENSOR_READ_PRESSURE;
      }
   else
   if (gnSensorState == SENSOR_READ_PRESSURE) {
      gnPraw = sns_ReadPressureSample();
      sns_TriggerTemperatureSample();
      tmr1_FlagWaitMs(13);
      gnSensorState = SENSOR_READ_TEMPERATURE;
      }
   }

void sns_TriggerPressureSample(void) {
   i2c_XmtByte(I2C_ID_BMP085,0xF4,0x34+(gnSensorOversampling<<6));
   }

void sns_TriggerTemperatureSample(void) {
   i2c_XmtByte(I2C_ID_BMP085,0xF4,0x2E);
   }

// Initialization of altitude sample buffer with average altitude
void sns_InitWindowBuffers(void) {
   s32 tc,tAccum,n,pa,pAccum;
   pAccum = 0;
   tAccum = 0;
   n = 8;
    while (n--) {
      sns_TriggerTemperatureSample();
      while(!gpioSnsEoc);
      gnTraw = sns_ReadTemperatureSample();
      tc = sns_CalcTemperatureCx10();
      sns_TriggerPressureSample();
      while(!gpioSnsEoc);
      gnPraw = sns_ReadPressureSample();
      pa = sns_CalcPressurePa();
      pAccum += pa;
      tAccum += tc;
      }
   tc = tAccum/8;
   gnTempC = (tc >= 0 ?  (tc+5)/10 : (tc-5)/10);
   gnPa = (pAccum+4)/8;
   gnAltAvgCm = sns_Pa2Cm(gnPa);
   gnAltAvgM = (gnAltAvgCm >= 0 ? (gnAltAvgCm+50L)/100L : (gnAltAvgCm-50L)/100L);
   n = SNS_NUM_Z_SAMPLES;
   while (n--) {
      gZBuf[n] = gnAltAvgCm;
      }
   gnSmpCnt = 0;
#ifdef DEBUG
   sprintf(gszBuf,"Tavg : %dC\r\n",gnTempC);
   uart_Print(gszBuf);
   sprintf(gszBuf,"Pavg : %ldPa\r\n",gnPa);
   uart_Print(gszBuf);
   sprintf(gszBuf,"Zavg : %ldm\r\n",gnAltAvgM);
   uart_Print(gszBuf);
#endif
   }

u32 sns_ReadTemperatureSample(void)	{
   u8 b[2];
   u32 w;
   i2c_RcvBuf(I2C_ID_BMP085,0xF6,2,b);
   w = (((u32)b[0])<<8) | (u32)b[1];
   return w;
   }

u32 sns_ReadPressureSample(void)	{
   u8 b[3];
   u32 dw;
   i2c_RcvBuf(I2C_ID_BMP085,0xF6,3,b);
   dw = (((u32)b[0])<<16) | (((u32)b[1])<<8) | (u32)b[2];
   dw = (dw >> (8-gnSensorOversampling));
   return dw;
   }

// These functions are provided by Bosch for converting the raw temperature and pressure A/D
// conversion data  into celsius and pascal readings
s16 sns_CalcTemperatureCx10(void)	{
   s32 x1,x2;
   s16 t;
   x1 = (((s32) gnTraw - (s32) gAC6) * (s32) gAC5) >> 15;
   x2 = ((s32) gMc << 11) / (x1 + gMd);
   gB5 = x1 + x2;
   t = ((gB5+8)>>4);
   return t;
   }

s32 sns_CalcPressurePa(void)	{
   s32 x1,x2,x3,b3,b6,p;
   u32 b4,b7;

   b6 = gB5 - 4000L;
   x1 = (b6*b6) >> 12;
   x1 *= gB2;
   x1 >>= 11;

   x2 = (gAC2*b6);
   x2 >>= 11;
   x3 = x1 +x2;
   b3 = (((((s32)gAC1 )*4 + x3) << gnSensorOversampling) + 2) >> 2;

   x1 = (gAC3* b6) >> 13;
   x2 = (gB1 * ((b6*b6) >> 12) ) >> 16;
   x3 = ((x1 + x2) + 2) >> 2;
   b4 = (gAC4 * (u32) (x3 + 32768L)) >> 15;

   b7 = ((u32)(gnPraw - b3) * (50000L >> gnSensorOversampling));   
   if (b7 < (u32)0x80000000) {
      p = (b7 << 1) / b4;
      }
   else { 
      p = (b7 / b4) << 1;
      }

   x1 = p >> 8;
   x1 *= x1;
   x1 = (x1 * 3038L) >> 16;
   x2 = (p * -7357L) >> 16;
   p += ((x1 + x2 + 3791L) >> 4);	// p in Pa  
   return p;
   }

void sns_AverageAltitude(void) {
   int cnt;
   s32 avg;
   cnt = SNS_NUM_Z_SAMPLES;
   avg = 0;
   while (cnt--)  {
	  avg += gZBuf[cnt];
      }
   avg = (avg >= 0 ? avg +SNS_NUM_Z_SAMPLES/2 : avg - SNS_NUM_Z_SAMPLES/2);
   gnAltAvgCm = avg/SNS_NUM_Z_SAMPLES;
   gnAltAvgM = ((gnAltAvgCm >= 0) ? (gnAltAvgCm+50L)/100L : (gnAltAvgCm-50L)/100L); 
   return; 
   }

// calc ok to +/- 20m/s with 32bit arithmetic
void sns_DeltaCmPerSec(void)   {
   int n,trel;
   s32 z, sum_zt,nCps;
   
   sum_zt = 0;
   n = SNS_NUM_Z_SAMPLES;
   while (n--)  {
      z = gZBuf[n] - gnAltAvgCm;
      trel = n - gnSmpCnt; // time origin is the current sample in window
      if (n > gnSmpCnt) {
         trel -= SNS_NUM_Z_SAMPLES;
         }
      sum_zt += ((s32)trel*z);
      }

   nCps = (sum_zt*(s32)(SENSOR_SAMPLES_PER_SEC*SNS_NUM_Z_SAMPLES))/gnDen;
   CLAMP(nCps,-2000,2000);
   gnAccel = (s16)(nCps - gnCps);
   sns_LPFAccel();
   gnCps = nCps;
   return;
   }

// Fast Lookup+Interpolation method for converting pressure readings to altitude readings.
#include "pztbl.txt"

s32 sns_Pa2Cm(s32 pa)  {
   	s32 inx,pa1,z1,z2,z;
    
   	if (pa > PA_INIT) {
      	z = gPZTbl[0];
      	}
   	else {
      	inx = (PA_INIT - pa)>>10;      
      	if (inx >= PZLUT_ENTRIES-1) {
         	z = gPZTbl[PZLUT_ENTRIES-1];
         	}
      	else {
         	pa1 = PA_INIT - (inx<<10);
         	z1 = gPZTbl[inx];
         	z2 = gPZTbl[inx+1];
         	z = z1 + (((pa1-pa)*(z2-z1))>>10);
         	}
      	}
   	return z;
   	}

#define I2C_ACK 		0
#define I2C_NAK 		1
#define I2C_WADDR(x) ((x) & 0xFE) //clear R/W bit of I2C addr
#define I2C_RADDR(x) ((x) | 0x01) //set R/W bit of I2C addr

// BMP085 allows max i2c clock of 3.4MHz, so use fastest i2c clock 
// possible (it cannot be higher than FCY anyway)
void i2c_Config(void) {
   //u32 temp;
   //temp = (FCY/1000L)/I2C_CLOCK_KHZ;
   //temp = temp - FCY/10000000L - 1;
   //I2C1BRG = temp;
   I2C1BRG = 0;
   I2C1CONbits.I2CEN = 1;
   }

void i2c_Start(void) {
   I2C1CONbits.SEN = 1;   // initiate start
   // wait until start finished
   while (I2C1CONbits.SEN);
   }

/*
void i2c_ReStart(void) {// repeated start
   I2C1CONbits.RSEN = 1;   // initiate start
   // wait until start finished
   while (I2C1CONbits.RSEN);
   }
*/

void i2c_Stop(void) {
   I2C1CONbits.PEN=1;     // initiate stop, PEN=1
   //wait until stop finished
   while (I2C1CONbits.PEN);
   }

void i2c_Put(u8 val) {
   I2C1TRN = val;    // write byte
   while (I2C1STATbits.TRSTAT);   // wait for 8bits+ ack bit to finish
   if (I2C1STATbits.ACKSTAT != I2C_ACK) {
      //NAK returned
      //uart_Print("I2CPUT1, NAK returned.");
      }
   }

/*
u8 i2c_PutNoAckCheck(u8 val) {
   I2C1TRN = val;    // write byte
   while (I2C1STATbits.TRSTAT);   // wait for 8bits+ ack bit to finish
   return(I2C1STATbits.ACKSTAT);
   }
*/

u8 i2c_Get(u8 ack2Send) {
   u8 inByte;
   while (I2C1CON & 0x1F);           //wait for idle condition
   I2C1CONbits.RCEN = 1;             //enable receive
   while (!I2C1STATbits.RBF);        //wait for receive byte
   inByte = I2C1RCV;              //read byte;
   //wait for idle condition before attempting ACK
   while (I2C1CON & 0x1F);           //lower 5 bits must be 0
   I2C1CONbits.ACKDT = ack2Send;  //ACK bit to send back on receive
   I2C1CONbits.ACKEN = 1;            //enable ACKbit transmittion
   while (I2C1CONbits.ACKEN);        //wait for completion
   return(inByte);                  //return the value
   }


void i2c_Write2(u8 addr,u8 d1, u8 d2) {
   i2c_Start();
   i2c_Put(I2C_WADDR( addr));
   i2c_Put( d1);
   i2c_Put( d2);
   i2c_Stop();
   }


void i2c_Write1(u8 addr,u8 d1) {
   i2c_Start();
   i2c_Put(I2C_WADDR(addr));
   i2c_Put(d1);
   i2c_Stop();
   }

/*

void i2c_WriteN(u8  addr,u8* pdata, u16  cnt) {
   u16  i;
   i2c_Start();
   i2c_Put(I2C_WADDR( addr));
   for ( i=0;  i <  cnt;  i++) {
      i2c_Put(*pdata);
      pdata++;
      }
   i2c_Stop();
   }


void i2c_Read1 (u8  addr,u8* pd1) {
   i2c_Start();
   i2c_Put(I2C_RADDR( addr));
   *pd1 = i2c_Get(I2C_NAK); //last ack bit from master to slave during read must be a NAK
   i2c_Stop();
   }


void i2c_Read2(u8  addr,u8* pd1, u8* pd2) {
   i2c_Start();
   i2c_Put(I2C_RADDR( addr));
   *pd1 = i2c_Get(I2C_ACK);
   *pd2 = i2c_Get(I2C_NAK);
   i2c_Stop();
   }
*/

void i2c_ReadN(u8  addr,u8* pdata, u16  cnt) {
   u16  inx;
   i2c_Start();
   i2c_Put(I2C_RADDR( addr));
   for (inx = 0;  inx <  cnt-1;  inx++) {
      *pdata++ = i2c_Get(I2C_ACK);
      }
   *pdata = i2c_Get(I2C_NAK);
   i2c_Stop();
   }

int i2c_XmtByte(u8 devId, u8 addr, u8 data) {
   i2c_Write2(devId,addr,data);
   return 0;
   }

int i2c_RcvBuf(u8 devId, u8 addr, int nBytes, u8* pBuf) {
   i2c_Write1(devId,addr);
   i2c_ReadN(devId, pBuf, nBytes);
   return 0;
   }

// TMR3 used for 50mS clock time tick
#define FREQ_TMR3_HZ   20L

void tmr3_Config(void){
	_T3IE = 0;  // disable interrupt
	T2CONbits.T32 = 0; // 16bit timer
    T3CONbits.TON = 0; // 16bit timer
    T3CONbits.TCS = 0; //  select internal clock 
    T3CONbits.TGATE = 0; // gate disabled
    T3CONbits.TSIDL = 0; // continue in idle mode
    T3CONbits.TCKPS1 = 0; // Prescaler 1:8
    T3CONbits.TCKPS0 = 1; // Prescaler 1:8
    PR3 = (u16)((FCY>>3)/FREQ_TMR3_HZ );
    gbTmr3Flag = 0;
    gnTimeTick = 0;
    gnSecTick = 0;
    gnElapsedSec = 0;
    gnElapsedMin = 0;
    gnElapsedHrs = 0;
    TMR3  = 0;   //zero timer3 counter
    _T3IF = 0;   //clear interrupt flag
    _T3IP = 1;   //choose a priority
	_T3IE = 1;  // enable interrupt
    T3CONbits.TON = 1; // start tmr3
	}

void __attribute__((interrupt, no_auto_psv)) _T3Interrupt(void) {
  	_T3IF = 0;    //clear the interrupt
   gbTmr3Flag = 1;
   gnTimeTick++;
   gnSecTick++;
   if (gnSecTick == FREQ_TMR3_HZ)  {
      gnSecTick = 0;
      gnElapsedSec++;
      if (gnElapsedSec == 60) {
         gnElapsedSec = 0;
         gnElapsedMin++;
         if (gnElapsedMin == 60)  {
            gnElapsedMin = 0;
            gnElapsedHrs++;
            }
         }      
      }
   }

// TMR1 used for busy wait, flag wait, elapsed time count
void  tmr1_Config(void) {
   T1CONbits.TON = 0;    // turn off the timer
   T1CONbits.TCS = 0;    // select internal clock 
   T1CONbits.TGATE = 0;  // gate disabled
   T1CONbits.TSIDL = 0;  // continue in idle mode
   T1CONbits.TCKPS1 = 1; // Prescaler 1:256
   T1CONbits.TCKPS0 = 1; // Prescaler 1:256
   gbTmr1Flag = 0;
   PR1 = 65535;
   TMR1  = 0;            //clear timer1 value
   _T1IF = 0;            //clear interrupt flag
   _T1IP = 1;            //choose a priority
   _T1IE = 0;            //disable the interrupt
   }

void tmr1_Start(void) {
   	T1CONbits.TON = 0; //turn off the timer
   	TMR1  = 0;         //zero timer1 counter
   	T1CONbits.TON = 1; //turn on the timer
	}

int tmr1_ElapsedTimeMs(void) {
    int ms;
    _T1IE = 0;            //disable the interrupt
   	T1CONbits.TON = 0; //turn off the timer
    ms = (int)(((TMR1*1000L)<<8)/FCY);
    return ms;
	}

// max delay is 67108 mS with FCY = 250kHz and prescaler=256
void tmr1_FlagWaitMs(u32 mS) {
    u16 cnt = (u16)(((mS * FCY)/1000L)>>8);
   	T1CONbits.TON = 0; // turn off timer
    gbTmr1Flag = 0;
   	TMR1 = 0;
    PR1  = cnt;         // set timer
   	T1CONbits.TON = 1; // turn on timer
    _T1IE = 1;            //enable the interrupt
	}    

void __attribute__((interrupt, no_auto_psv)) _T1Interrupt(void) {
   _T1IF = 0;    //clear the timer interrupt bit
   T1CONbits.TON = 0; // turn off timer
   _T1IE = 0;            //disable the interrupt   
   gbTmr1Flag = 1;
   }    

// max delay is 67108 mS with FCY = 250kHz and prescaler=256
void tmr1_BusyWaitMs(u32 mS) {
    u16 cnt = (u16)(((mS * FCY)/1000L)>>8);
   	T1CONbits.TON = 0; // turn off timer
   	TMR1  = 0;         // zero timer1 counter
   	T1CONbits.TON = 1; // turn on timer
    while(TMR1 < cnt); // busy wait until interval completes
    T1CONbits.TON = 0; // turn off timer again
    return;
	}

#ifdef DEBUG

void uart_Config(void) {
   char tmp;
   U2BRG  = 6;     // 8928baud @ FCY = 250kHz, note that you must select "custom baud rate" in the
                   // pickit2 uart application drop-down window for baudrate, and enter the value 8928                   
   U2MODEbits.BRGH = 1;
   U2MODEbits.UARTEN = 1;
   U2MODEbits.RTSMD = 1; // no flow control
   U2MODEbits.PDSEL = 0; // 8 bits, no parity
   U2MODEbits.STSEL = 0; // 1 stop bit
   while(U2STAbits.URXDA == 1) {
      tmp = U2RXREG;
      }
   // clear IF flags 
   IFS1bits.U2RXIF = 0;
   IFS1bits.U2TXIF = 0;
   // set priority 
   IPC7bits.U2RXIP = 1;
   IPC7bits.U2TXIP = 1;
   // disable interrupt 
   IEC1bits.U2RXIE = 0;
   IEC1bits.U2TXIE = 0;
   // enable transmit
   U2STAbits.UTXEN = 1;
   }

void uart_Print(char *szBuf) {
   while(*szBuf) {
      U2TXREG = *szBuf++;   // transfer data byte to TX reg 
      while(!U2STAbits.TRMT);  // wait for xmit to complete
      }
   }

// Test to check the Bosch provided code that converts the sensor raw data 
// to actual pressure in Pa and temperature in Celsius.
// Check that the code produces the same results as in the pResult and tResult arrays.

u16 ac5[9] = {32758, 32758, 32758, 30774, 30774, 30774, 26431, 26431, 26431};
u16 ac6[9] = {23153, 23153, 23153, 9469, 9469, 9469, 40099, 40099, 40099};

u32 up[9] = {25827, 23843, 21476, 25634, 23656, 21302, 26060, 24056, 21657};
u32 ut[9] = {24524, 27898, 34787, 9389, 14520, 21853, 40438, 45980, 54518};

s32  pResult[9]= {70369, 69964, 69961, 66140, 69406, 69385, 68346, 70599, 70559 };
s16  tResult[9] = {-177, 150, 650, -404, 150, 650, -338, 150, 650};

void sns_BoschCodeTest(void) {
   int cnt,bResult;
   s16 t;
   s32 p;
   gAC1 = 408;
   gAC2 = -72;
   gAC3 = -14383;
   gAC4 = 32741;
   gB1 = 6190;
   gB2 = 4;
   gMb = 0;
   gMc = -8711;
   gMd = 2868;
   gnSensorOversampling = 0L; // no oversampling while running bosch unit test
   bResult = 0;
   for (cnt = 0; cnt < 9; cnt++){
      gAC5 = ac5[cnt];
      gAC6 = ac6[cnt];
      gnTraw = ut[cnt];
      gnPraw = up[cnt];
      t = sns_CalcTemperatureCx10();
      p = sns_CalcPressurePa();
      bResult = (tResult[cnt] == t) && (pResult[cnt] == p);
      if (!bResult) break;
      }
   uart_Print(bResult ? "Bosch test passed\r\n" : "BOSCH TEST FAILED\r\n");
   gnSensorOversampling = 3L; // reset it to normal oversampling
   }

#endif // DEBUG
